package com.perforce.hws.reporting;

import org.jacoco.core.analysis.Analyzer;
import org.jacoco.core.analysis.CoverageBuilder;
import org.jacoco.core.analysis.IBundleCoverage;
import org.jacoco.core.tools.ExecFileLoader;
import org.jacoco.report.DirectorySourceFileLocator;
import org.jacoco.report.FileMultiReportOutput;
import org.jacoco.report.IReportVisitor;
import org.jacoco.report.html.HTMLFormatter;

import com.perforce.hws.automation.ProjectLocations;

import java.io.File;
import java.io.IOException;
import java.io.UncheckedIOException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.Arrays;
import java.util.List;

/**
 * Reads coverage data and creates our coverage report assuming our typical project structure.
 */
public class CreateCoverageReport implements ProjectLocations {

    public static void main(String args[]) {
        try {
            new CreateCoverageReport().run();
        } catch (Exception e) {
            e.printStackTrace();
            System.exit(1);
        }
    }

    private ExecFileLoader execFileLoader = new ExecFileLoader();

    private void run() throws IOException {
        loadAllExecutionData();

        final IBundleCoverage bundleCoverage = analyzeStructure();

        createReport(bundleCoverage);
    }

    private void createReport(IBundleCoverage coverage) throws IOException {
        HTMLFormatter htmlFormatter = new HTMLFormatter();

        Path reportDir = getReportDir();
        if (!Files.exists(reportDir)) {
            Files.createDirectories(reportDir);
        }

        IReportVisitor reportVisitor = htmlFormatter.createVisitor(new FileMultiReportOutput(reportDir.toFile()));

        reportVisitor.visitInfo(execFileLoader.getSessionInfoStore().getInfos(),
                execFileLoader.getExecutionDataStore().getContents());

        visitAllSources(reportVisitor, coverage);

        reportVisitor.visitEnd();
    }

    private void visitAllSources(IReportVisitor reportVisitor, IBundleCoverage coverage) {
        allSourcePaths().stream()
                .forEach(s -> {
                    try {
                        reportVisitor.visitBundle(coverage, new DirectorySourceFileLocator(s, "utf-8", 4));
                    } catch (IOException e) {
                        throw new UncheckedIOException(e);
                    }
                });
    }

    private void loadAllExecutionData() throws IOException {
        Files.list(Paths.get(coverageOutputPath()))
                .forEach(this::loadExecutionPath);
    }

    private void loadExecutionPath(Path file) {
        try {
            execFileLoader.load(file.toFile());
        } catch (IOException e) {
            throw new UncheckedIOException(e);
        }
    }

    private IBundleCoverage analyzeStructure() throws IOException {
        final CoverageBuilder coverageBuilder = new CoverageBuilder();
        final Analyzer analyzer = new Analyzer(
                execFileLoader.getExecutionDataStore(), coverageBuilder);

        analyzeAllClassDirectories(analyzer);

        return coverageBuilder.getBundle("Coverage");
    }

    private void analyzeAllClassDirectories(Analyzer analyzer) {
        allClassPaths().stream()
                .forEach(f -> {
                    try {
                        analyzer.analyzeAll(f);
                    } catch (IOException e) {
                        throw new UncheckedIOException(e);
                    }
                });
    }

    private List<File> allClassPaths() {
        try {
            return Arrays.asList(
                    Paths.get(projectRootOrFail(), "git_fusion", "build", "classes").toAbsolutePath().toFile(),
                    Paths.get(projectRootOrFail(), "jdk_extensions", "build", "classes").toAbsolutePath().toFile(),
                    Paths.get(projectRootOrFail(), "p4base", "build", "classes").toAbsolutePath().toFile(),
                    Paths.get(projectRootOrFail(), "server", "build", "classes").toAbsolutePath().toFile()
            );
        } catch (IOException e) {
            throw new UncheckedIOException(e);
        }
    }

    private List<File> allSourcePaths() {
        try {
            return Arrays.asList(
                    Paths.get(projectRootOrFail(), "git_fusion", "src", "main", "java").toAbsolutePath().toFile(),
                    Paths.get(projectRootOrFail(), "jdk_extensions", "src", "main", "java").toAbsolutePath().toFile(),
                    Paths.get(projectRootOrFail(), "p4base", "src", "main", "java").toAbsolutePath().toFile(),
                    Paths.get(projectRootOrFail(), "server", "src", "main", "java").toAbsolutePath().toFile()
            );
        } catch (IOException e) {
            throw new UncheckedIOException(e);
        }
    }


    private Path getReportDir(){
        try {
            return Paths.get(projectRootOrFail(), "reporting", "build", "coverageReport");
        } catch (IOException e) {
            throw new UncheckedIOException(e);
        }
    }
}
